@@ -1,4 +1,5 @@
 # pylint: disable=too-many-lines
+import re
 import json
 from collections import OrderedDict
 
@@ -97,6 +98,16 @@ class RemoteState(object):
             function_arn=deployed_values['lambda_arn'],
         )
 
+    def _resource_exists_lambdalayer(self, resource):
+        # type: (models.LambdaLayer) -> bool
+        try:
+            deployed_values = self._deployed_resources.resource_values(
+                resource.resource_name)
+        except ValueError:
+            return False
+        return bool(self._client.get_layer_version(
+            deployed_values['layer_version_arn']))
+
     def _resource_exists_lambdafunction(self, resource):
         # type: (models.LambdaFunction) -> bool
         return self._client.lambda_function_exists(resource.function_name)
@@ -347,6 +358,48 @@ class PlanStage(object):
         ])
         return api_calls
 
+    def _plan_lambdalayer(self, resource):
+        # type: (models.LambdaLayer) -> Sequence[InstructionMsg]
+
+        api_calls = []  # type: List[InstructionMsg]
+        filename = cast(str, resource.deployment_package.filename)
+
+        # Automatically clean up old layer versions.
+        # See:
+        # https://docs.aws.amazon.com/lambda/latest/dg/API_DeleteLayerVersion.html
+        msg = 'Creating'
+        if self._remote_state.resource_exists(resource):
+            state = self._remote_state.resource_deployed_values(resource)
+            # Deleting a layer version won't break functions still using it.
+            # From the doc link above:
+            #
+            # "To avoid breaking functions, a copy of the version remains in
+            # Lambda until no functions refer to it."
+            api_calls.append(
+                models.APICall(
+                    method_name='delete_layer_version',
+                    params={'layer_version_arn': state['layer_version_arn']}
+                )
+            )
+            msg = 'Updating'
+
+        api_calls.extend([(
+            models.APICall(
+                method_name='publish_layer',
+                params={'layer_name': resource.layer_name,
+                        'zip_contents': self._osutils.get_file_contents(
+                            filename, binary=True),
+                        'runtime': resource.runtime},
+                output_var='layer_version_arn'
+            ), "%s lambda layer: %s\n" % (msg, resource.layer_name)),
+            models.RecordResourceVariable(
+                resource_type='lambda_layer',
+                resource_name=resource.resource_name,
+                name='layer_version_arn',
+                variable_name='layer_version_arn',
+        )])
+        return api_calls
+
     def _plan_lambdafunction(self, resource):
         # type: (models.LambdaFunction) -> Sequence[InstructionMsg]
         role_arn = self._get_role_arn(resource.role)
@@ -362,7 +415,9 @@ class PlanStage(object):
         if resource.reserved_concurrency is None:
             concurrency_api_call = models.APICall(
                 method_name='delete_function_concurrency',
-                params={'function_name': resource.function_name},
+                params={
+                    'function_name': resource.function_name,
+                },
                 output_var='reserved_concurrency_result'
             )
         else:
@@ -380,6 +435,12 @@ class PlanStage(object):
             )
 
         api_calls = []  # type: List[InstructionMsg]
+        layers = []  # type: List[Any]
+        if resource.managed_layer is not None:
+            layers.append(Variable('layer_version_arn'))
+        if resource.layers:
+            layers.extend(resource.layers)
+
         if not self._remote_state.resource_exists(resource):
             params = {
                 'function_name': resource.function_name,
@@ -394,8 +455,9 @@ class PlanStage(object):
                 'memory_size': resource.memory_size,
                 'security_group_ids': resource.security_group_ids,
                 'subnet_ids': resource.subnet_ids,
-                'layers': resource.layers
+                'layers': layers
             }
+
             api_calls.extend([
                 (models.APICall(
                     method_name='create_function',
@@ -424,7 +486,7 @@ class PlanStage(object):
                 'memory_size': resource.memory_size,
                 'security_group_ids': resource.security_group_ids,
                 'subnet_ids': resource.subnet_ids,
-                'layers': resource.layers
+                'layers': layers
             }
             api_calls.extend([
                 (models.APICall(
@@ -454,10 +516,37 @@ class PlanStage(object):
         varname = '%s_role_arn' % resource.role_name
         if not role_exists:
             return [
+                models.BuiltinFunction(
+                    'service_principal',
+                    ['lambda'],
+                    output_var='lambda_service_principal',
+                ),
+                models.JPSearch('principal',
+                                input_var='lambda_service_principal',
+                                output_var='lambda_principal'),
+                models.StoreValue(
+                    name='lambda_principal',
+                    value=StringFormat('{lambda_principal}',
+                                       ['lambda_principal']),
+                ),
+                models.StoreValue(
+                    name='lambda_trust_policy',
+                    value={
+                        "Version": "2012-10-17",
+                        "Statement": [{
+                            "Sid": "",
+                            "Effect": "Allow",
+                            "Principal": {
+                                "Service": Variable('lambda_principal')
+                            },
+                            "Action": "sts:AssumeRole"
+                        }]
+                    },
+                ),
                 (models.APICall(
                     method_name='create_role',
                     params={'name': resource.role_name,
-                            'trust_policy': resource.trust_policy,
+                            'trust_policy': Variable('lambda_trust_policy'),
                             'policy': document},
                     output_var=varname,
                 ), "Creating IAM role: %s\n" % resource.role_name),
@@ -507,7 +596,7 @@ class PlanStage(object):
         subscribe_varname = '%s_subscription_arn' % resource.resource_name
 
         instruction_for_topic_arn = []  # type: List[InstructionMsg]
-        if resource.topic.startswith('arn:aws:sns:'):
+        if re.match(r"^arn:aws[a-z\-]*:sns:", resource.topic):
             instruction_for_topic_arn += [
                 models.StoreValue(
                     name=topic_arn_varname,
@@ -530,13 +619,16 @@ class PlanStage(object):
                 models.JPSearch('region',
                                 input_var='parsed_lambda_arn',
                                 output_var='region_name'),
+                models.JPSearch('partition',
+                                input_var='parsed_lambda_arn',
+                                output_var='partition'),
                 models.StoreValue(
                     name=topic_arn_varname,
                     value=StringFormat(
-                        'arn:aws:sns:{region_name}:{account_id}:%s' % (
+                        'arn:{partition}:sns:{region_name}:{account_id}:%s' % (
                             resource.topic
                         ),
-                        ['region_name', 'account_id'],
+                        ['partition', 'region_name', 'account_id'],
                     ),
                 ),
             ]
@@ -633,13 +725,16 @@ class PlanStage(object):
             models.JPSearch('region',
                             input_var='parsed_lambda_arn',
                             output_var='region_name'),
+            models.JPSearch('partition',
+                            input_var='parsed_lambda_arn',
+                            output_var='partition'),
             models.StoreValue(
                 name=queue_arn_varname,
                 value=StringFormat(
-                    'arn:aws:sqs:{region_name}:{account_id}:%s' % (
+                    'arn:{partition}:sqs:{region_name}:{account_id}:%s' % (
                         resource.queue
                     ),
-                    ['region_name', 'account_id'],
+                    ['partition', 'region_name', 'account_id'],
                 ),
             ),
         ]  # type: List[InstructionMsg]
@@ -834,11 +929,11 @@ class PlanStage(object):
                 models.StoreValue(
                     name='websocket-%s-integration-lambda-path' % key,
                     value=StringFormat(
-                        'arn:aws:apigateway:{region_name}:lambda:path/'
-                        '2015-03-31/functions/arn:aws:lambda:{region_name}:'
-                        '{account_id}:function:%s/'
-                        'invocations' % config['name'],
-                        ['region_name', 'account_id'],
+                        'arn:{partition}:apigateway:{region_name}:lambda:path/'
+                        '2015-03-31/functions/arn:{partition}'
+                        ':lambda:{region_name}:{account_id}:function'
+                        ':%s/invocations' % config['name'],
+                        ['partition', 'region_name', 'account_id'],
                     ),
                 ),
             )
@@ -895,6 +990,12 @@ class PlanStage(object):
             models.JPSearch('region',
                             input_var='parsed_lambda_arn',
                             output_var='region_name'),
+            models.JPSearch('partition',
+                            input_var='parsed_lambda_arn',
+                            output_var='partition'),
+            models.JPSearch('dns_suffix',
+                            input_var='parsed_lambda_arn',
+                            output_var='dns_suffix'),
         ]  # type: List[InstructionMsg]
 
         # There's also a set of instructions that are needed
@@ -905,8 +1006,8 @@ class PlanStage(object):
                 name='websocket_api_url',
                 value=StringFormat(
                     'wss://{websocket_api_id}.execute-api.{region_name}'
-                    '.amazonaws.com/%s/' % resource.api_gateway_stage,
-                    ['websocket_api_id', 'region_name'],
+                    '.{dns_suffix}/%s/' % resource.api_gateway_stage,
+                    ['websocket_api_id', 'region_name', 'dns_suffix'],
                 ),
             ),
             models.RecordResourceVariable(
@@ -1047,6 +1148,12 @@ class PlanStage(object):
             models.JPSearch('region',
                             input_var='parsed_lambda_arn',
                             output_var='region_name'),
+            models.JPSearch('partition',
+                            input_var='parsed_lambda_arn',
+                            output_var='partition'),
+            models.JPSearch('dns_suffix',
+                            input_var='parsed_lambda_arn',
+                            output_var='dns_suffix'),
             # The swagger doc uses the 'api_handler_lambda_arn'
             # var name so we need to make sure we populate this variable
             # before importing the rest API.
@@ -1086,8 +1193,8 @@ class PlanStage(object):
                 name='rest_api_url',
                 value=StringFormat(
                     'https://{rest_api_id}.execute-api.{region_name}'
-                    '.amazonaws.com/%s/' % resource.api_gateway_stage,
-                    ['rest_api_id', 'region_name'],
+                    '.{dns_suffix}/%s/' % resource.api_gateway_stage,
+                    ['rest_api_id', 'region_name', 'dns_suffix'],
                 ),
             ),
             models.RecordResourceVariable(
