@@ -3,10 +3,12 @@
 import copy
 import json
 import os
+import re
 
 import six
 from typing import Any, Optional, Dict, List, Set, Union  # noqa
 from typing import cast
+
 import yaml
 from yaml.scanner import ScannerError
 from yaml.nodes import Node, ScalarNode, SequenceNode
@@ -16,6 +18,7 @@ from chalice.deploy.swagger import (
 from chalice.utils import (
     OSUtils, UI, serialize_to_json, to_cfn_resource_name
 )
+from chalice.awsclient import TypedAWSClient # noqa
 from chalice.config import Config  # noqa
 from chalice.deploy import models
 from chalice.deploy.appgraph import ApplicationGraphBuilder, DependencyBuilder
@@ -24,9 +27,9 @@ from chalice.deploy.deployer import create_build_stage
 
 
 def create_app_packager(
-        config, package_format='cloudformation',
+        config, options, package_format='cloudformation',
         template_format='json', merge_template=None):
-    # type: (Config, str, str, Optional[str]) -> AppPackager
+    # type: (Config, PackageOptions, str, str, Optional[str]) -> AppPackager
     osutils = OSUtils()
     ui = UI()
     application_builder = ApplicationGraphBuilder()
@@ -37,7 +40,7 @@ def create_app_packager(
     template_serializer = cast(TemplateSerializer, JSONTemplateSerializer())
     if package_format == 'cloudformation':
         build_stage = create_build_stage(
-            osutils, ui, CFNSwaggerGenerator())
+            osutils, ui, CFNSwaggerGenerator(), config)
         use_yaml_serializer = template_format == 'yaml'
         if merge_template is not None and \
                 YAMLTemplateSerializer.is_yaml_template(merge_template):
@@ -54,11 +57,11 @@ def create_app_packager(
                 merger=TemplateDeepMerger(),
                 template_serializer=template_serializer,
                 merge_template=merge_template)])
-        generator = SAMTemplateGenerator(config)
+        generator = SAMTemplateGenerator(config, options)
     else:
         build_stage = create_build_stage(
-            osutils, ui, TerraformSwaggerGenerator())
-        generator = TerraformGenerator(config)
+            osutils, ui, TerraformSwaggerGenerator(), config)
+        generator = TerraformGenerator(config, options)
         post_processors.append(
             TerraformCodeLocationPostProcessor(osutils=osutils))
 
@@ -81,6 +84,20 @@ class DuplicateResourceNameError(Exception):
     pass
 
 
+class PackageOptions(object):
+    def __init__(self, client):
+        # type: (TypedAWSClient) -> None
+        self._client = client  # type: TypedAWSClient
+
+    def service_principal(self, service):
+        # type: (str) -> str
+        dns_suffix = self._client.endpoint_dns_suffix(service,
+                                                      self._client.region_name)
+        return self._client.service_principal(service,
+                                              self._client.region_name,
+                                              dns_suffix)
+
+
 class ResourceBuilder(object):
     def __init__(self,
                  application_builder,  # type: ApplicationGraphBuilder
@@ -98,6 +115,9 @@ class ResourceBuilder(object):
             config, chalice_stage_name)
         resources = self._deps_builder.build_dependencies(application)
         self._build_stage.execute(config, resources)
+        # Rebuild dependencies in case the build stage modified
+        # the application graph.
+        resources = self._deps_builder.build_dependencies(application)
         return resources
 
 
@@ -105,9 +125,10 @@ class TemplateGenerator(object):
 
     template_file = None  # type: str
 
-    def __init__(self, config):
-        # type: (Config) -> None
+    def __init__(self, config, options):
+        # type: (Config, PackageOptions) -> None
         self._config = config
+        self._options = options
 
     def dispatch(self, resource, template):
         # type: (models.Model, Dict[str, Any]) -> None
@@ -141,7 +162,6 @@ class TemplateGenerator(object):
 
 
 class SAMTemplateGenerator(TemplateGenerator):
-
     _BASE_TEMPLATE = {
         'AWSTemplateFormatVersion': '2010-09-09',
         'Transform': 'AWS::Serverless-2016-10-31',
@@ -151,10 +171,11 @@ class SAMTemplateGenerator(TemplateGenerator):
 
     template_file = "sam"
 
-    def __init__(self, config):
-        # type: (Config) -> None
-        super(SAMTemplateGenerator, self).__init__(config)
+    def __init__(self, config, options):
+        # type: (Config, PackageOptions) -> None
+        super(SAMTemplateGenerator, self).__init__(config, options)
         self._seen_names = set([])  # type: Set[str]
+        self._chalice_layer = ""
 
     def generate(self, resources):
         # type: (List[models.Model]) -> Dict[str, Any]
@@ -164,6 +185,20 @@ class SAMTemplateGenerator(TemplateGenerator):
             self.dispatch(resource, template)
         return template
 
+    def _generate_lambdalayer(self, resource, template):
+        # type: (models.LambdaLayer, Dict[str, Any]) -> None
+        layer = to_cfn_resource_name(
+            resource.resource_name)
+        template['Resources'][layer] = {
+            "Type": "AWS::Serverless::LayerVersion",
+            "Properties": {
+                "CompatibleRuntimes": [resource.runtime],
+                "ContentUri": resource.deployment_package.filename,
+                "LayerName": resource.layer_name
+            }
+        }
+        self._chalice_layer = layer
+
     def _generate_scheduledevent(self, resource, template):
         # type: (models.ScheduledEvent, Dict[str, Any]) -> None
         function_cfn_name = to_cfn_resource_name(
@@ -235,10 +270,15 @@ class SAMTemplateGenerator(TemplateGenerator):
             }
             lambdafunction_definition['Properties'].update(
                 reserved_concurrency_config)
-        if resource.layers:
+
+        layers = resource.layers or []  # type: List[Any]
+        if self._chalice_layer:
+            layers.insert(0, {'Ref': self._chalice_layer})
+
+        if layers:
             layers_config = {
-                'Layers': resource.layers
-            }  # type: Dict[str, List[str]]
+                'Layers': layers
+            }  # type: Dict[str, Any]
             lambdafunction_definition['Properties'].update(layers_config)
 
         resources[cfn_name] = lambdafunction_definition
@@ -284,11 +324,11 @@ class SAMTemplateGenerator(TemplateGenerator):
             'Properties': {
                 'FunctionName': {'Ref': 'APIHandler'},
                 'Action': 'lambda:InvokeFunction',
-                'Principal': 'apigateway.amazonaws.com',
+                'Principal': self._options.service_principal('apigateway'),
                 'SourceArn': {
                     'Fn::Sub': [
-                        ('arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}'
-                         ':${RestAPIId}/*'),
+                        ('arn:${AWS::Partition}:execute-api:${AWS::Region}'
+                         ':${AWS::AccountId}:${RestAPIId}/*'),
                         {'RestAPIId': {'Ref': 'RestAPI'}},
                     ]
                 },
@@ -301,11 +341,12 @@ class SAMTemplateGenerator(TemplateGenerator):
                 'Properties': {
                     'FunctionName': {'Fn::GetAtt': [auth_cfn_name, 'Arn']},
                     'Action': 'lambda:InvokeFunction',
-                    'Principal': 'apigateway.amazonaws.com',
+                    'Principal': self._options.service_principal('apigateway'),
                     'SourceArn': {
                         'Fn::Sub': [
-                            ('arn:aws:execute-api:${AWS::Region}:'
-                             '${AWS::AccountId}:${RestAPIId}/*'),
+                            ('arn:${AWS::Partition}:execute-api'
+                             ':${AWS::Region}:${AWS::AccountId}'
+                             ':${RestAPIId}/*'),
                             {'RestAPIId': {'Ref': 'RestAPI'}},
                         ]
                     },
@@ -343,7 +384,7 @@ class SAMTemplateGenerator(TemplateGenerator):
                     'https://${RestAPI}.execute-api.${AWS::Region}'
                     # The api_gateway_stage is filled in when
                     # the template is built.
-                    '.amazonaws.com/%s/'
+                    '.${AWS::URLSuffix}/%s/'
                 ) % stage_name
             }
         }
@@ -361,10 +402,11 @@ class SAMTemplateGenerator(TemplateGenerator):
                 'IntegrationUri': {
                     'Fn::Sub': [
                         (
-                            'arn:aws:apigateway:${AWS::Region}:lambda:path/'
-                            '2015-03-31/functions/arn:aws:lambda:'
-                            '${AWS::Region}:' '${AWS::AccountId}:function:'
-                            '${WebsocketHandler}/invocations'
+                            'arn:${AWS::Partition}:apigateway:${AWS::Region}'
+                            ':lambda:path/2015-03-31/functions/arn'
+                            ':${AWS::Partition}:lambda:${AWS::Region}'
+                            ':${AWS::AccountId}:function'
+                            ':${WebsocketHandler}/invocations'
                         ),
                         {'WebsocketHandler': {'Ref': websocket_handler}}
                     ],
@@ -380,10 +422,11 @@ class SAMTemplateGenerator(TemplateGenerator):
             'Properties': {
                 'FunctionName': {'Ref': websocket_handler},
                 'Action': 'lambda:InvokeFunction',
-                'Principal': 'apigateway.amazonaws.com',
+                'Principal': self._options.service_principal('apigateway'),
                 'SourceArn': {
                     'Fn::Sub': [
-                        ('arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}'
+                        ('arn:${AWS::Partition}:execute-api'
+                         ':${AWS::Region}:${AWS::AccountId}'
                          ':${WebsocketAPIId}/*'),
                         {'WebsocketAPIId': api_ref},
                     ],
@@ -516,7 +559,7 @@ class SAMTemplateGenerator(TemplateGenerator):
                     'wss://${WebsocketAPI}.execute-api.${AWS::Region}'
                     # The api_gateway_stage is filled in when
                     # the template is built.
-                    '.amazonaws.com/%s/'
+                    '.${AWS::URLSuffix}/%s/'
                 ) % stage_name
             }
         }
@@ -530,6 +573,8 @@ class SAMTemplateGenerator(TemplateGenerator):
         # type: (models.ManagedIAMRole, Dict[str, Any]) -> None
         role_cfn_name = self._register_cfn_resource_name(
             resource.resource_name)
+        resource.trust_policy['Statement'][0]['Principal']['Service'] = \
+            self._options.service_principal('lambda')
         template['Resources'][role_cfn_name] = {
             'Type': 'AWS::IAM::Role',
             'Properties': {
@@ -559,12 +604,13 @@ class SAMTemplateGenerator(TemplateGenerator):
         sns_cfn_name = self._register_cfn_resource_name(
             resource.resource_name)
 
-        if resource.topic.startswith('arn:aws:sns:'):
+        if re.match(r"^arn:aws[a-z\-]*:sns:", resource.topic):
             topic_arn = resource.topic  # type: Union[str, Dict[str, str]]
         else:
             topic_arn = {
                 'Fn::Sub': (
-                    'arn:aws:sns:${AWS::Region}:${AWS::AccountId}:%s' %
+                    'arn:${AWS::Partition}:sns'
+                    ':${AWS::Region}:${AWS::AccountId}:%s' %
                     resource.topic
                 )
             }
@@ -590,7 +636,8 @@ class SAMTemplateGenerator(TemplateGenerator):
                 'Properties': {
                     'Queue': {
                         'Fn::Sub': (
-                            'arn:aws:sqs:${AWS::Region}:${AWS::AccountId}:%s' %
+                            'arn:${AWS::Partition}:sqs:${AWS::Region}'
+                            ':${AWS::AccountId}:%s' %
                             resource.queue
                         )
                     },
@@ -687,9 +734,13 @@ class SAMTemplateGenerator(TemplateGenerator):
 
 
 class TerraformGenerator(TemplateGenerator):
-
     template_file = "chalice.tf"
 
+    def __init__(self, config, options):
+        # type: (Config, PackageOptions) -> None
+        super(TerraformGenerator, self).__init__(config, options)
+        self._chalice_layer = ""
+
     def generate(self, resources):
         # type: (List[models.Model]) -> Dict[str, Any]
         template = {
@@ -704,6 +755,7 @@ class TerraformGenerator(TemplateGenerator):
             },
             'data': {
                 'aws_caller_identity': {'chalice': {}},
+                'aws_partition': {'chalice': {}},
                 'aws_region': {'chalice': {}},
                 'null_data_source': {
                     'chalice': {
@@ -728,6 +780,7 @@ class TerraformGenerator(TemplateGenerator):
     def _arnref(self, arn_template, **kw):
         # type: (str, str) -> str
         d = dict(
+            partition='${data.aws_partition.chalice.partition}',
             region='${data.aws_region.chalice.name}',
             account_id='${data.aws_caller_identity.chalice.account_id}')
         d.update(kw)
@@ -735,17 +788,21 @@ class TerraformGenerator(TemplateGenerator):
 
     def _generate_managediamrole(self, resource, template):
         # type: (models.ManagedIAMRole, Dict[str, Any]) -> None
+
+        resource.trust_policy['Statement'][0]['Principal']['Service'] = \
+            self._options.service_principal('lambda')
+
         template['resource'].setdefault('aws_iam_role', {})[
             resource.resource_name] = {
-                'name': resource.role_name,
-                'assume_role_policy': json.dumps(resource.trust_policy)
+            'name': resource.role_name,
+            'assume_role_policy': json.dumps(resource.trust_policy)
         }
 
         template['resource'].setdefault('aws_iam_role_policy', {})[
             resource.resource_name] = {
-                'name': resource.resource_name + 'Policy',
-                'policy': json.dumps(resource.policy.document),
-                'role': '${aws_iam_role.%s.id}' % resource.resource_name,
+            'name': resource.resource_name + 'Policy',
+            'policy': json.dumps(resource.policy.document),
+            'role': '${aws_iam_role.%s.id}' % resource.resource_name,
         }
 
     def _generate_websocketapi(self, resource, template):
@@ -783,28 +840,29 @@ class TerraformGenerator(TemplateGenerator):
             bucket_name = resource.bucket
         template['resource'].setdefault(
             'aws_s3_bucket_notification', {}).setdefault(
-                bucket_name + '_notify',
-                {'bucket': resource.bucket}).setdefault(
-                    'lambda_function', []).append(bnotify)
+            bucket_name + '_notify',
+            {'bucket': resource.bucket}).setdefault(
+            'lambda_function', []).append(bnotify)
 
         template['resource'].setdefault('aws_lambda_permission', {})[
             resource.resource_name] = {
-                'statement_id': resource.resource_name,
-                'action': 'lambda:InvokeFunction',
-                'function_name': resource.lambda_function.function_name,
-                'principal': 's3.amazonaws.com',
-                'source_arn': 'arn:aws:s3:::%s' % resource.bucket
+            'statement_id': resource.resource_name,
+            'action': 'lambda:InvokeFunction',
+            'function_name': resource.lambda_function.function_name,
+            'principal': self._options.service_principal('s3'),
+            'source_arn': 'arn:*:s3:::%s' % resource.bucket
         }
 
     def _generate_sqseventsource(self, resource, template):
         # type: (models.SQSEventSource, Dict[str, Any]) -> None
         template['resource'].setdefault('aws_lambda_event_source_mapping', {})[
             resource.resource_name] = {
-                'event_source_arn': self._arnref(
-                    "arn:aws:sqs:%(region)s:%(account_id)s:%(queue)s",
-                    queue=resource.queue),
-                'batch_size': resource.batch_size,
-                'function_name': resource.lambda_function.function_name,
+            'event_source_arn': self._arnref(
+                "arn:%(partition)s:sqs:%(region)s"
+                ":%(account_id)s:%(queue)s",
+                queue=resource.queue),
+            'batch_size': resource.batch_size,
+            'function_name': resource.lambda_function.function_name,
         }
 
     def _generate_snslambdasubscription(self, resource, template):
@@ -814,21 +872,21 @@ class TerraformGenerator(TemplateGenerator):
             topic_arn = resource.topic
         else:
             topic_arn = self._arnref(
-                'arn:aws:sns:%(region)s:%(account_id)s:%(topic)s',
+                'arn:%(partition)s:sns:%(region)s:%(account_id)s:%(topic)s',
                 topic=resource.topic)
 
         template['resource'].setdefault('aws_sns_topic_subscription', {})[
             resource.resource_name] = {
-                'topic_arn': topic_arn,
-                'protocol': 'lambda',
-                'endpoint': self._fref(resource.lambda_function)
+            'topic_arn': topic_arn,
+            'protocol': 'lambda',
+            'endpoint': self._fref(resource.lambda_function)
         }
         template['resource'].setdefault('aws_lambda_permission', {})[
             resource.resource_name] = {
-                'function_name': resource.lambda_function.function_name,
-                'action': 'lambda:InvokeFunction',
-                'principal': 'sns.amazonaws.com',
-                'source_arn': topic_arn
+            'function_name': resource.lambda_function.function_name,
+            'action': 'lambda:InvokeFunction',
+            'principal': self._options.service_principal('sns'),
+            'source_arn': topic_arn
         }
 
     def _generate_cloudwatchevent(self, resource, template):
@@ -836,9 +894,9 @@ class TerraformGenerator(TemplateGenerator):
 
         template['resource'].setdefault(
             'aws_cloudwatch_event_rule', {})[
-                resource.resource_name] = {
-                    'name': resource.resource_name,
-                    'event_pattern': resource.event_pattern
+            resource.resource_name] = {
+            'name': resource.resource_name,
+            'event_pattern': resource.event_pattern
         }
         self._cwe_helper(resource, template)
 
@@ -847,10 +905,10 @@ class TerraformGenerator(TemplateGenerator):
 
         template['resource'].setdefault(
             'aws_cloudwatch_event_rule', {})[
-                resource.resource_name] = {
-                    'name': resource.resource_name,
-                    'schedule_expression': resource.schedule_expression,
-                    'description': resource.rule_description,
+            resource.resource_name] = {
+            'name': resource.resource_name,
+            'schedule_expression': resource.schedule_expression,
+            'description': resource.rule_description,
         }
         self._cwe_helper(resource, template)
 
@@ -858,25 +916,35 @@ class TerraformGenerator(TemplateGenerator):
         # type: (models.CloudWatchEventBase, Dict[str, Any]) -> None
         template['resource'].setdefault(
             'aws_cloudwatch_event_target', {})[
-                resource.resource_name] = {
-                    'rule': '${aws_cloudwatch_event_rule.%s.name}' % (
-                        resource.resource_name),
-                    'target_id': resource.resource_name,
-                    'arn': self._fref(resource.lambda_function)
+            resource.resource_name] = {
+            'rule': '${aws_cloudwatch_event_rule.%s.name}' % (
+                resource.resource_name),
+            'target_id': resource.resource_name,
+            'arn': self._fref(resource.lambda_function)
         }
         template['resource'].setdefault(
             'aws_lambda_permission', {})[
-                resource.resource_name] = {
-                    'function_name': resource.lambda_function.function_name,
-                    'action': 'lambda:InvokeFunction',
-                    'principal': 'events.amazonaws.com',
-                    'source_arn': "${aws_cloudwatch_event_rule.%s.arn}" % (
-                        resource.resource_name)
+            resource.resource_name] = {
+            'function_name': resource.lambda_function.function_name,
+            'action': 'lambda:InvokeFunction',
+            'principal': self._options.service_principal('events'),
+            'source_arn': "${aws_cloudwatch_event_rule.%s.arn}" % (
+                resource.resource_name)
         }
 
+    def _generate_lambdalayer(self, resource, template):
+        # type: (models.LambdaLayer, Dict[str, Any]) -> None
+        template['resource'].setdefault(
+            "aws_lambda_layer_version", {})[
+                resource.resource_name] = {
+                'layer_name': resource.layer_name,
+                'compatible_runtimes': [resource.runtime],
+                'filename': resource.deployment_package.filename,
+        }
+        self._chalice_layer = resource.resource_name
+
     def _generate_lambdafunction(self, resource, template):
         # type: (models.LambdaFunction, Dict[str, Any]) -> None
-
         func_definition = {
             'function_name': resource.function_name,
             'runtime': resource.runtime,
@@ -886,7 +954,8 @@ class TerraformGenerator(TemplateGenerator):
             'timeout': resource.timeout,
             'source_code_hash': '${filebase64sha256("%s")}' % (
                 resource.deployment_package.filename),
-            'filename': resource.deployment_package.filename}
+            'filename': resource.deployment_package.filename
+        }  # type: Dict[str, Any]
 
         if resource.security_group_ids and resource.subnet_ids:
             func_definition['vpc_config'] = {
@@ -901,8 +970,13 @@ class TerraformGenerator(TemplateGenerator):
             func_definition['environment'] = {
                 'variables': resource.environment_variables
             }
+        if self._chalice_layer:
+            func_definition['layers'] = [
+                '${aws_lambda_layer_version.%s.arn}' % self._chalice_layer
+            ]
         if resource.layers:
-            func_definition['layers'] = list(resource.layers)
+            func_definition.setdefault('layers', []).extend(
+                list(resource.layers))
 
         if isinstance(resource.role, models.ManagedIAMRole):
             func_definition['role'] = '${aws_iam_role.%s.arn}' % (
@@ -922,72 +996,78 @@ class TerraformGenerator(TemplateGenerator):
         swagger_doc = cast(Dict, resource.swagger_doc)
         template['data'].setdefault(
             'template_file', {}).setdefault(
-                'chalice_api_swagger', {})['template'] = json.dumps(
-                    swagger_doc)
+            'chalice_api_swagger', {})['template'] = json.dumps(
+            swagger_doc)
 
         template['resource'].setdefault('aws_api_gateway_rest_api', {})[
             resource.resource_name] = {
-                'body': '${data.template_file.chalice_api_swagger.rendered}',
-                # Terraform will diff explicitly configured attributes
-                # to the current state of the resource. Attributes configured
-                # via swagger on the REST api need to be duplicated here, else
-                # terraform will set them back to empty.
-                'name': swagger_doc['info']['title'],
-                'binary_media_types': swagger_doc[
-                    'x-amazon-apigateway-binary-media-types'],
-                'endpoint_configuration': {'types': [resource.endpoint_type]}
+            'body': '${data.template_file.chalice_api_swagger.rendered}',
+            # Terraform will diff explicitly configured attributes
+            # to the current state of the resource. Attributes configured
+            # via swagger on the REST api need to be duplicated here, else
+            # terraform will set them back to empty.
+            'name': swagger_doc['info']['title'],
+            'binary_media_types': swagger_doc[
+                'x-amazon-apigateway-binary-media-types'],
+            'endpoint_configuration': {'types': [resource.endpoint_type]}
         }
 
         if 'x-amazon-apigateway-policy' in swagger_doc:
             template['resource'][
                 'aws_api_gateway_rest_api'][
-                    resource.resource_name]['policy'] = json.dumps(
-                        swagger_doc['x-amazon-apigateway-policy'])
+                resource.resource_name]['policy'] = json.dumps(
+                swagger_doc['x-amazon-apigateway-policy'])
         if resource.minimum_compression.isdigit():
             template['resource'][
                 'aws_api_gateway_rest_api'][
-                    resource.resource_name][
-                        'minimum_compression_size'] = int(
-                            resource.minimum_compression)
+                resource.resource_name][
+                'minimum_compression_size'] = int(
+                resource.minimum_compression)
 
         template['resource'].setdefault('aws_api_gateway_deployment', {})[
             resource.resource_name] = {
-                'stage_name': resource.api_gateway_stage,
-                # Ensure that the deployment gets redeployed if we update
-                # the swagger description for the api by using its checksum
-                # in the stage description.
-                'stage_description': (
-                    "${md5(data.template_file.chalice_api_swagger.rendered)}"),
-                'rest_api_id': '${aws_api_gateway_rest_api.%s.id}' % (
-                    resource.resource_name),
-                'lifecycle': {'create_before_destroy': True}
+            'stage_name': resource.api_gateway_stage,
+            # Ensure that the deployment gets redeployed if we update
+            # the swagger description for the api by using its checksum
+            # in the stage description.
+            'stage_description': (
+                "${md5(data.template_file.chalice_api_swagger.rendered)}"),
+            'rest_api_id': '${aws_api_gateway_rest_api.%s.id}' % (
+                resource.resource_name),
+            'lifecycle': {'create_before_destroy': True}
         }
 
         template['resource'].setdefault('aws_lambda_permission', {})[
             resource.resource_name + '_invoke'] = {
-                'function_name': resource.lambda_function.function_name,
-                'action': 'lambda:InvokeFunction',
-                'principal': 'apigateway.amazonaws.com',
-                'source_arn':
-                    "${aws_api_gateway_rest_api.%s.execution_arn}/*" % (
-                        resource.resource_name)
+            'function_name': resource.lambda_function.function_name,
+            'action': 'lambda:InvokeFunction',
+            'principal': self._options.service_principal('apigateway'),
+            'source_arn':
+                "${aws_api_gateway_rest_api.%s.execution_arn}/*" % (
+                    resource.resource_name)
         }
 
         template.setdefault('output', {})[
             'EndpointURL'] = {
-                'value': '${aws_api_gateway_deployment.%s.invoke_url}' % (
-                    resource.resource_name)
+            'value': '${aws_api_gateway_deployment.%s.invoke_url}' % (
+                resource.resource_name)
+        }
+        template.setdefault('output', {})[
+            'RestAPIId'] = {
+            'value': '${aws_api_gateway_rest_api.%s.id}' % (
+                resource.resource_name)
         }
 
         for auth in resource.authorizers:
             template['resource']['aws_lambda_permission'][
                 auth.resource_name + '_invoke'] = {
-                    'function_name': auth.function_name,
-                    'action': 'lambda:InvokeFunction',
-                    'principal': 'apigateway.amazonaws.com',
-                    'source_arn': (
-                        "${aws_api_gateway_rest_api.%s.execution_arn}" % (
-                            resource.resource_name) + "/*")
+                'function_name': auth.function_name,
+                'action': 'lambda:InvokeFunction',
+                'principal': self._options.service_principal('apigateway'),
+                'source_arn': (
+                    "${aws_api_gateway_rest_api.%s.execution_arn}" % (
+                        resource.resource_name) + "/*"
+                )
             }
         self._add_domain_name(resource, template)
 
@@ -1033,11 +1113,11 @@ class TerraformGenerator(TemplateGenerator):
 
 class AppPackager(object):
     def __init__(self,
-                 templater,            # type: TemplateGenerator
-                 resource_builder,     # type: ResourceBuilder
-                 post_processor,       # type: TemplatePostProcessor
+                 templater,  # type: TemplateGenerator
+                 resource_builder,  # type: ResourceBuilder
+                 post_processor,  # type: TemplatePostProcessor
                  template_serializer,  # type: TemplateSerializer
-                 osutils,              # type: OSUtils
+                 osutils,  # type: OSUtils
                  ):
         # type: (...) -> None
         self._templater = templater
@@ -1102,14 +1182,18 @@ class SAMCodeLocationPostProcessor(TemplatePostProcessor):
         # somehow, which isn't currently possible.
         copied = False
         for resource in template['Resources'].values():
-            if resource['Type'] != 'AWS::Serverless::Function':
-                continue
-            original_location = resource['Properties']['CodeUri']
-            new_location = os.path.join(outdir, 'deployment.zip')
-            if not copied:
+            if resource['Type'] == 'AWS::Serverless::Function':
+                original_location = resource['Properties']['CodeUri']
+                new_location = os.path.join(outdir, 'deployment.zip')
+                if not copied:
+                    self._osutils.copy(original_location, new_location)
+                    copied = True
+                resource['Properties']['CodeUri'] = './deployment.zip'
+            elif resource['Type'] == 'AWS::Serverless::LayerVersion':
+                original_location = resource['Properties']['ContentUri']
+                new_location = os.path.join(outdir, 'layer-deployment.zip')
                 self._osutils.copy(original_location, new_location)
-                copied = True
-            resource['Properties']['CodeUri'] = './deployment.zip'
+                resource['Properties']['ContentUri'] = './layer-deployment.zip'
 
 
 class TerraformCodeLocationPostProcessor(TemplatePostProcessor):
@@ -1118,7 +1202,8 @@ class TerraformCodeLocationPostProcessor(TemplatePostProcessor):
         # type: (Dict[str, Any], Config, str, str) -> None
 
         copied = False
-        for r in template['resource'].get('aws_lambda_function', {}).values():
+        resources = template['resource']
+        for r in resources.get('aws_lambda_function', {}).values():
             if not copied:
                 asset_path = os.path.join(outdir, 'deployment.zip')
                 self._osutils.copy(r['filename'], asset_path)
@@ -1126,12 +1211,21 @@ class TerraformCodeLocationPostProcessor(TemplatePostProcessor):
             r['filename'] = "${path.module}/deployment.zip"
             r['source_code_hash'] = \
                 '${filebase64sha256("${path.module}/deployment.zip")}'
+        copied = False
+        for r in resources.get('aws_lambda_layer_version', {}).values():
+            if not copied:
+                asset_path = os.path.join(outdir, 'layer-deployment.zip')
+                self._osutils.copy(r['filename'], asset_path)
+                copied = True
+            r['filename'] = "${path.module}/layer-deployment.zip"
+            r['source_code_hash'] = \
+                '${filebase64sha256("${path.module}/layer-deployment.zip")}'
 
 
 class TemplateMergePostProcessor(TemplatePostProcessor):
     def __init__(self,
-                 osutils,              # type: OSUtils
-                 merger,               # type: TemplateMerger
+                 osutils,  # type: OSUtils
+                 merger,  # type: TemplateMerger
                  template_serializer,  # type: TemplateSerializer
                  merge_template=None,  # type: Optional[str]
                  ):
@@ -1187,7 +1281,7 @@ class TemplateDeepMerger(TemplateMerger):
     def _merge(self, file_template, chalice_template):
         # type: (Any, Any) -> Any
         if isinstance(file_template, dict) and \
-           isinstance(chalice_template, dict):
+                isinstance(chalice_template, dict):
             return self._merge_dict(file_template, chalice_template)
         return file_template
 
@@ -1200,7 +1294,6 @@ class TemplateDeepMerger(TemplateMerger):
 
 
 class TemplateSerializer(object):
-
     file_extension = ''
 
     def load_template(self, file_contents, filename=''):
@@ -1213,7 +1306,6 @@ class TemplateSerializer(object):
 
 
 class JSONTemplateSerializer(TemplateSerializer):
-
     file_extension = 'json'
 
     def serialize_template(self, contents):
@@ -1230,7 +1322,6 @@ class JSONTemplateSerializer(TemplateSerializer):
 
 
 class YAMLTemplateSerializer(TemplateSerializer):
-
     file_extension = 'yaml'
 
     @classmethod
