@@ -7,6 +7,7 @@ from collections import defaultdict, namedtuple
 import pytest
 import mock
 
+from chalice.awsclient import TypedAWSClient
 from chalice.config import Config
 from chalice import Chalice
 from chalice import package
@@ -19,6 +20,7 @@ from chalice.deploy.packager import SDistMetadataFetcher
 from chalice.deploy.packager import InvalidSourceDistributionNameError
 from chalice.compat import pip_no_compile_c_env_vars
 from chalice.compat import pip_no_compile_c_shim
+from chalice.package import PackageOptions
 from chalice.utils import OSUtils
 
 
@@ -925,7 +927,7 @@ class TestDependencyBuilder(object):
         assert installed_packages == ['bar']
 
 
-def test_can_create_app_packager_with_no_autogen(tmpdir):
+def test_can_create_app_packager_with_no_autogen(tmpdir, stubbed_session):
     appdir = _create_app_structure(tmpdir)
 
     outdir = tmpdir.mkdir('outdir')
@@ -933,7 +935,8 @@ def test_can_create_app_packager_with_no_autogen(tmpdir):
     config = Config.create(project_dir=str(appdir),
                            chalice_app=sample_app(),
                            **default_params)
-    p = package.create_app_packager(config)
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options)
     p.package_app(config, str(outdir), 'dev')
     # We're not concerned with the contents of the files
     # (those are tested in the unit tests), we just want to make
@@ -943,7 +946,7 @@ def test_can_create_app_packager_with_no_autogen(tmpdir):
     assert 'sam.json' in contents
 
 
-def test_can_create_app_packager_with_yaml_extention(tmpdir):
+def test_can_create_app_packager_with_yaml_extention(tmpdir, stubbed_session):
     appdir = _create_app_structure(tmpdir)
 
     outdir = tmpdir.mkdir('outdir')
@@ -953,7 +956,9 @@ def test_can_create_app_packager_with_yaml_extention(tmpdir):
     config = Config.create(project_dir=str(appdir),
                            chalice_app=sample_app(),
                            **default_params)
-    p = package.create_app_packager(config, merge_template=str(extras_file))
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options,
+                                    merge_template=str(extras_file))
 
     p.package_app(config, str(outdir), 'dev')
     contents = os.listdir(str(outdir))
@@ -961,7 +966,7 @@ def test_can_create_app_packager_with_yaml_extention(tmpdir):
     assert 'sam.yaml' in contents
 
 
-def test_can_specify_yaml_output(tmpdir):
+def test_can_specify_yaml_output(tmpdir, stubbed_session):
     appdir = _create_app_structure(tmpdir)
 
     outdir = tmpdir.mkdir('outdir')
@@ -969,7 +974,8 @@ def test_can_specify_yaml_output(tmpdir):
     config = Config.create(project_dir=str(appdir),
                            chalice_app=sample_app(),
                            **default_params)
-    p = package.create_app_packager(config, template_format='yaml')
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options, template_format='yaml')
 
     p.package_app(config, str(outdir), 'dev')
     contents = os.listdir(str(outdir))
@@ -977,20 +983,58 @@ def test_can_specify_yaml_output(tmpdir):
     assert 'sam.yaml' in contents
 
 
-def test_will_create_outdir_if_needed(tmpdir):
+def test_will_create_outdir_if_needed(tmpdir, stubbed_session):
     appdir = _create_app_structure(tmpdir)
     outdir = str(appdir.join('outdir'))
     default_params = {'autogen_policy': True}
     config = Config.create(project_dir=str(appdir),
                            chalice_app=sample_app(),
                            **default_params)
-    p = package.create_app_packager(config)
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options)
     p.package_app(config, str(outdir), 'dev')
     contents = os.listdir(str(outdir))
     assert 'deployment.zip' in contents
     assert 'sam.json' in contents
 
 
+def test_includes_layer_package_with_sam(tmpdir, stubbed_session):
+    appdir = _create_app_structure(tmpdir)
+    appdir.mkdir('vendor').join('hello').write('hello\n')
+    outdir = str(appdir.join('outdir'))
+    default_params = {'autogen_policy': True}
+    config = Config.create(project_dir=str(appdir),
+                           chalice_app=sample_app(),
+                           automatic_layer=True,
+                           **default_params)
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options)
+    p.package_app(config, str(outdir), 'dev')
+    contents = os.listdir(str(outdir))
+    assert 'deployment.zip' in contents
+    assert 'layer-deployment.zip' in contents
+    assert 'sam.json' in contents
+
+
+def test_includes_layer_package_with_terraform(tmpdir, stubbed_session):
+    appdir = _create_app_structure(tmpdir)
+    appdir.mkdir('vendor').join('hello').write('hello\n')
+    outdir = str(appdir.join('outdir'))
+    default_params = {'autogen_policy': True}
+    config = Config.create(project_dir=str(appdir),
+                           chalice_app=sample_app(),
+                           automatic_layer=True,
+                           **default_params)
+    options = PackageOptions(TypedAWSClient(session=stubbed_session))
+    p = package.create_app_packager(config, options,
+                                    package_format='terraform')
+    p.package_app(config, str(outdir), 'dev')
+    contents = os.listdir(str(outdir))
+    assert 'deployment.zip' in contents
+    assert 'layer-deployment.zip' in contents
+    assert 'chalice.tf.json' in contents
+
+
 class TestSubprocessPip(object):
     def test_can_invoke_pip(self):
         pip = SubprocessPip()
